==> angular_components-0.5.1.tar.gz: lib/src/model/selection/string_selection_options.dart <==
// Copyright (c) 2016, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

library angular_components.model.selection.string_selection_options;

import 'package:meta/meta.dart';

import '../../utils/async/async.dart';
import '../ui/has_renderer.dart';
import './select.dart';
import './selection_options.dart';

/// Format string to be insensitive on case and spaces.
String _stringFormatSuggestion/*<T>*/(/*=T*/ Object value) =>
    '$value'.replaceAll(' ', '').toLowerCase();

typedef StringSuggestionFilter(suggestion, String filterQuery);

/// The class is meant to be used in areas where a selection can be
/// represented as a string.
/// Ex:
/// class Color {
///   final int colorId;
///   final String displayName;
///
///   Color(this.colorId, this.displayName);
/// }
///
/// var options = new StringSelectionOptions<Color>(
///     // list of colors
///     [new Color (1, 'Red'), new Color(2, 'Blue'), new Color(3, Purple)],
///     // converts color object to string for filtering.
///     (Color color) => color.displayName.toLowerCase());
///
class StringSelectionOptions<T> extends SelectionOptions<T>
    implements Filterable {
  /// Unlimited large value to support no limit for filtering.
  static const int UNLIMITED = 9007199254740992;

  /// Store the last query passed to filter.
  String _currentQuery;

  /// The current limit for the filter being applied.
  int _currentLimit = -1;

  /// List of options.
  List<OptionGroup<T>> _optionGroups;

  /// The itemRenderer is used for filtering the options.
  ItemRenderer<T> _toFilterableString;

  /// Function for filtering a single suggestion/option, defaults to
  /// [filterOption] method.
  StringSuggestionFilter _suggestionFilter;

  bool _shouldSort = false;

  /// The list of options and optionally a function to convert the option
  /// into a string that can be used for filtering the list.
  ///
  /// The options must be either strings or convertible to strings using an
  /// optional [ItemRenderer] that is passed in.  If no [ItemRenderer] is
  /// passed in, [_stringFormatSuggestion] will be used.
  ///
  /// If the data needs to be sorted, it should be passed in sorted, options
  /// .sort() is a simple way apply the default sorting rules.
  StringSelectionOptions(List<T> options,
      {ItemRenderer<T> toFilterableString,
      StringSuggestionFilter suggestionFilter,
      bool shouldSort: false})
      : this.withOptionGroups([new OptionGroup(options)],
            toFilterableString: toFilterableString,
            suggestionFilter: suggestionFilter,
            shouldSort: shouldSort);

  StringSelectionOptions.withOptionGroups(List<OptionGroup<T>> optionGroups,
      {ItemRenderer<T> toFilterableString,
      StringSuggestionFilter suggestionFilter,
      bool shouldSort: false})
      : _toFilterableString = toFilterableString ??
            new CachingItemRenderer<T>(_stringFormatSuggestion).call,
        _shouldSort = shouldSort,
        super(optionGroups) {
    _suggestionFilter =
        suggestionFilter != null ? suggestionFilter : filterOption;
  }

  /// Accepts a string query and limit and applies the filter to the options.
  ///
  /// If the limit is 0 or negative, it will return all possible options
  /// that match the query.
  ///
  /// TODO(google): In a followup CL, refactor the reusable portions of this code
  /// into a generic filter.
  ///
  @override
  DisposableFuture<bool> filter(Object query, {int limit: -1}) {
    _currentLimit = limit < 1 ? UNLIMITED : limit;
    _currentQuery = query as String;
    refilter();
    return new DisposableFuture.fromValue(true);
  }

  @protected
  bool get filterApplied => currentQuery != null;

  @protected
  void refilter() {
    List<OptionGroup<T>> filtered = [];
    int count = 0;
    String filterQuery =
        _currentQuery == null ? '' : _stringFormatSuggestion(_currentQuery);
    for (var group in _optionGroups) {
      if (count >= currentLimit) break;
      var filteredGroup =
          filterOptionGroup(group, filterQuery, currentLimit - count);
      count += filteredGroup.length;
      filtered.add(filteredGroup);
    }
    super.optionGroups = filtered;
  }

  /// Filter a single OptionGroup.
  /// Method is exposed to allow override.
  OptionGroup<T> filterOptionGroup(
      OptionGroup<T> group, String filterQuery, int limit) {
    Iterable<T> list;
    if (filterQuery.isNotEmpty) {
      // TODO(google): Optimize the code below to stop filtering if the
      // limit is met rather than filtering everything than taking the
      // matches.
      list = group
          .where((suggestion) => _suggestionFilter(suggestion, filterQuery))
          .take(limit);
    } else {
      list = group.take(limit);
    }
    var filteredGroup = new OptionGroup<T>.withLabelFunction(
        list.toList(growable: false),
        () => group.uiDisplayName,
        group.emptyLabel != null ? () => group.emptyLabel : null);

    return filteredGroup;
  }

  /// Filters a single option.
  /// Method is exposed to allow override.
  bool filterOption(T option, String filterQuery) {
    // StringFormatSuggestion is used to eliminate spaces to make the
    // pattern matching a simple contains as opposed to a regex.
    return _stringFormatSuggestion(_toFilterableString(option))
        .contains(filterQuery);
  }

  @override
  set optionGroups(List<OptionGroup<T>> value) {
    // This mutates value...
    if (_shouldSort) {
      value.forEach((optionGroup) {
        optionGroup.sort(_sortFn);
      });
    }
    // Cache the data so original is kept during filtering.
    _optionGroups = value;
    super.optionGroups = value;
    if (filterApplied) {
      refilter();
    }
  }

  int _sortFn(T a, T b) =>
      _toFilterableString(a).compareTo(_toFilterableString(b));

  @override
  Object get currentQuery => _currentQuery;

  @override
  int get currentLimit => _currentLimit;

  /// Returns the unfiltered list of option groups.
  List<OptionGroup<T>> get unfilteredOptionGroups => _optionGroups;
}

==> angular_components-0.7.0.tar.gz: lib/src/model/selection/string_selection_options.dart <==
// Copyright (c) 2016, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

import 'package:meta/meta.dart';

import '../../utils/async/async.dart';
import '../ui/has_renderer.dart';
import './select.dart';
import './selection_options.dart';

/// Formats [value] as a lowercase string without spaces.
String _stringFormatSuggestion<T>(T value) =>
    '$value'.replaceAll(' ', '').toLowerCase();

typedef StringSuggestionFilter(suggestion, String filterQuery);

/// The class is meant to be used in areas where a selection can be represented
/// as a string.
///
/// __Example usage:__
/// ```dart
/// class Color {
///   final int colorId;
///   final String displayName;
///
///   Color(this.colorId, this.displayName);
/// }
///
/// var options = new StringSelectionOptions<Color>(
///     // list of colors
///     [new Color (1, 'Red'), new Color(2, 'Blue'), new Color(3, 'Purple')],
///     // converts color object to string for filtering.
///     (Color color) => color.displayName.toLowerCase());
/// ```
class StringSelectionOptions<T> extends SelectionOptions<T>
    implements Filterable {
  /// Unlimited large value to support no limit for filtering.
  static const int UNLIMITED = 9007199254740992;

  /// The last query passed to [filter].
  String _currentQuery;

  /// The current limit for the filter being applied.
  int _currentLimit = -1;

  List<OptionGroup<T>> _optionGroups;

  /// A function that converts a single option to a filterable string.
  ItemRenderer<T> _toFilterableString;

  /// Function for filtering a single suggestion/option, defaults to
  /// [filterOption] method.
  StringSuggestionFilter _suggestionFilter;

  /// The [ItemRenderer] that sanitizes options and queries before filtering.
  ItemRenderer _sanitizeString;

  bool _shouldSort = false;

  /// The list of options and optionally a function to convert the option into a
  /// string that can be used for filtering the list.
  ///
  /// The options must be either strings or convertible to strings using an
  /// optional [ItemRenderer] that is passed in.  If no [ItemRenderer] is passed
  /// in, [_stringFormatSuggestion] will be used.
  ///
  /// If the data needs to be sorted, it should be passed in sorted. The option
  /// [shouldSort] is a simple way to apply the default sorting rules.
  StringSelectionOptions(List<T> options,
      {ItemRenderer<T> toFilterableString,
      StringSuggestionFilter suggestionFilter,
      ItemRenderer sanitizeString: _stringFormatSuggestion,
      bool shouldSort: false})
      : this.withOptionGroups([new OptionGroup(options)],
            toFilterableString: toFilterableString,
            suggestionFilter: suggestionFilter,
            sanitizeString: sanitizeString,
            shouldSort: shouldSort);

  StringSelectionOptions.withOptionGroups(List<OptionGroup<T>> optionGroups,
      {ItemRenderer<T> toFilterableString,
      StringSuggestionFilter suggestionFilter,
      ItemRenderer sanitizeString: _stringFormatSuggestion,
      bool shouldSort: false})
      : _toFilterableString = toFilterableString ??
            new CachingItemRenderer<T>(sanitizeString).call,
        _shouldSort = shouldSort,
        super(optionGroups) {
    _suggestionFilter =
        suggestionFilter != null ? suggestionFilter : filterOption;
    _sanitizeString = sanitizeString;
  }

  /// Accepts a string query and limit and applies the filter to the options.
  ///
  /// If the limit is 0 or negative, it will return all possible options that
  /// match the query.
  ///
  /// TODO(google): In a followup CL, refactor the reusable portions of this code
  /// into a generic filter.
  @override
  DisposableFuture<bool> filter(Object query, {int limit: -1}) {
    _currentLimit = limit < 1 ? UNLIMITED : limit;
    _currentQuery = query as String;
    refilter();
    return new DisposableFuture.fromValue(true);
  }

  @protected
  bool get filterApplied => currentQuery != null;

  @protected
  void refilter() {
    List<OptionGroup<T>> filtered = [];
    int count = 0;
    String filterQuery =
        _currentQuery == null ? '' : _sanitizeString(_currentQuery);
    for (var group in _optionGroups) {
      if (count >= currentLimit) break;
      var filteredGroup =
          filterOptionGroup(group, filterQuery, currentLimit - count);
      count += filteredGroup.length;
      filtered.add(filteredGroup);
    }
    super.optionGroups = filtered;
  }

  @protected
  OptionGroup<T> filterOptionGroup(
      OptionGroup<T> group, String filterQuery, int limit) {
    Iterable<T> list;
    if (filterQuery.isNotEmpty) {
      // TODO(google): Optimize the code below to stop filtering if the
      // limit is met rather than filtering everything than taking the
      // matches.
      list = group
          .where((suggestion) => _suggestionFilter(suggestion, filterQuery))
          .take(limit);
    } else {
      list = group.take(limit);
    }
    var filteredGroup = new OptionGroup<T>.withLabelFunction(
        list.toList(growable: false),
        () => group.uiDisplayName,
        group.emptyLabel != null ? () => group.emptyLabel : null);

    return filteredGroup;
  }

  @protected
  bool filterOption(T option, String filterQuery) {
    // StringFormatSuggestion is used to eliminate spaces to make the
    // pattern matching a simple contains as opposed to a regex.
    return _sanitizeString(_toFilterableString(option)).contains(filterQuery);
  }

  @override
  Object get currentQuery => _currentQuery;

  @override
  int get currentLimit => _currentLimit;

  List<OptionGroup<T>> get unfilteredOptionGroups => _optionGroups;

  @override
  set optionGroups(List<OptionGroup<T>> value) {
    // This mutates value...
    if (_shouldSort) {
      value.forEach((optionGroup) {
        optionGroup.sort(_sortFn);
      });
    }
    // Cache the data so original is kept during filtering.
    _optionGroups = value;
    super.optionGroups = value;
    if (filterApplied) {
      refilter();
    }
  }

  int _sortFn(T a, T b) =>
      _toFilterableString(a).compareTo(_toFilterableString(b));
}
